#include <failsystem.h>

#ifdef FATMAPPED
static u_int *fat;
#endif

#ifdef MTHREAD
static pthread_mutex_t fatmux=PTHREAD_MUTEX_INITIALIZER,fdmux=PTHREAD_MUTEX_INITIALIZER;
#endif

/*this function return a file descriptor to the filesystem file*/
int fsys_open(char *filename){
	return open(filename,O_RDWR);
}

/*close file of given fd*/
int fsys_close(int fd){
	return close(fd);
}

/*read superblock structure from given fd*/
int fsys_sblockget(int fd,struct sblock *sb){
/*	pthread_mutex_lock(&fdmux);*/
	lseek(fd,0,SEEK_SET);
	read(fd,(void*)sb,sizeof(struct sblock));
/*	pthread_mutex_unlock(&fdmux);*/
	return 0;
}

/*return the next linked block*/
inline u_int fsys_getnextblock(int fd,struct sblock *sb,u_int index){
	return fsys_fatread(fd,sb,index);
}

/* add one or more blocks to the free list */
int fsys_blockfree(int fd,struct sblock *sb,p_block index){
p_block current,nexttofree;	
	if (!index){return -1;}
	current=index;
	while(current){
		nexttofree=fsys_getnextblock(fd,sb,current);
		fsys_fatwrite(fd,sb,current,sb->firstfree);
		sb->firstfree=current;
		current=nexttofree;
		fsys_printlog(FSYS_LOG_SYSLOG,"blockfree::BLOCK FREED IS: %u\n",sb->firstfree);
	}

	return 0;
}

/*return a new block */
u_int fsys_blocknew(int fd,struct sblock *sb){
u_int ret;
	if (!sb->firstfree){
		return FSYS_DISKFULL;
	}
	ret=sb->firstfree;
	sb->firstfree=fsys_getnextblock(fd,sb,ret);
	fsys_printlog(FSYS_LOG_SYSLOG,"blocknew::BLOCK NEW IS: %u\nBLOCK FREE IS: %u\n",ret,sb->firstfree);
	fsys_fatwrite(fd,sb,ret,FSYS_BUSY);
	return ret;
}

/*return the tail of a list of blocks*/
u_int fsys_getlastblock(int fd,struct sblock *sb,u_int index){
u_int current;
	current=(-1);
	while (index){
		current=index;
		index=fsys_getnextblock(fd,sb,index);
	}
	return current;
}

/*links and allocate a new block to given block*/
u_int fsys_blockappend(int fd,struct sblock *sb,u_int blocktoappendto){
u_int ret;
	ret=fsys_blocknew(fd,sb);
	if (ret==FSYS_DISKFULL) return ret;
	fsys_fatwrite(fd,sb,blocktoappendto,ret);
	return ret;
}

/*writes a entry in the fat array*/
int fsys_fatwrite(int fd,struct sblock *sb,u_int index,u_int data){
	fsys_checkbounds(sb,index);
#ifdef FATMAPPED	
	fat[index]=data;
	return 0;
#else
u_int offset;
	offset= sb->blocksize + (index * sizeof(u_int));
/*	pthread_mutex_lock(&fdmux);*/
	lseek(fd,offset,SEEK_SET);
	write(fd,(void*)&data,sizeof(u_int));
/*	pthread_mutex_unlock(&fdmux);*/
	return 0;
#endif
}

/*reads a entry from fat*/
u_int fsys_fatread(int fd,struct sblock *sb,u_int index){
	fsys_checkbounds(sb,index);
#ifdef FATMAPPED
	return fat[index];
#else
u_int offset,ret;
	offset= sb->blocksize + (index * sizeof(u_int));
/*	pthread_mutex_lock(&fdmux);*/
	lseek(fd,offset,SEEK_SET);
	read(fd,(void*)&ret,sizeof(u_int));
/*	pthread_mutex_unlock(&fdmux);*/
	return ret;
	
#endif
}

/*read data from a block*/
int fsys_blockread(int fd,struct sblock *sb,u_int block,void *buf,u_int offset, size_t size){
u_int bread;	
	fsys_checkbounds(sb,block);
	if ((offset + size) > sb->blocksize){
		return -1;
	}
/*	pthread_mutex_lock(&fdmux);*/
	lseek(fd,(sb->blocksize * block) + offset,SEEK_SET);
	bread=read(fd,buf,size);
/*	pthread_mutex_unlock(&fdmux);*/
	return bread;
}

/*writes data to a block*/
int fsys_blockwrite(int fd,struct sblock *sb,u_int block,void *buf,u_int offset, size_t size){
u_int bwritten;	
	fsys_checkbounds(sb,block);
	if ((offset + size) > sb->blocksize){
		return -1;
	}
/*	pthread_mutex_lock(&fdmux);*/
	lseek(fd,(sb->blocksize * block) + offset,SEEK_SET);
	bwritten=write(fd,buf,size);
/*	pthread_mutex_unlock(&fdmux);*/
	return bwritten;
}

/*loads fat into memory*/
#ifdef FATMAPPED	
int fsys_fatload(int fd,struct sblock *sb){
	fat=(u_int*)malloc(sb->fatsize);
	memset(fat,0,sb->fatsize);
	lseek(fd,sb->blocksize,SEEK_SET);
	return read(fd,fat,sb->fatsize);
}
#endif

/* sync fat stored into disk from the content of array loaded in memory */
#ifdef FATMAPPED	
int fsys_fatsync(int fd,struct sblock *sb){
/*	pthread_mutex_lock(&fdmux);*/
	lseek(fd,sb->blocksize,SEEK_SET);
	write(fd,fat,sb->fatsize);
/*	pthread_mutex_unlock(&fdmux);*/
	return 0;
}
#endif

/* sanity check for fat access */
int fsys_checkbounds(struct sblock *sb,p_block index){
	if (index > sb->blockscount){	/*check sanity: index must be between FAT bounds*/
		fsys_printlog(FSYS_LOG_SYSLOG,"FAT index %u out of range!\n",index);
		fsys_printlog(FSYS_LOG_STDERR,"Exiting...");
		exit (1);
		return	0;
	}
	return 0;
}
